import pytest
import responses

from darwin.future.core.client import ClientCore
from darwin.future.core.items.move_items_to_folder import move_list_of_items_to_folder
from darwin.future.exceptions import BadRequest
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_move_list_of_items_to_folder_including_filters(
    base_client: ClientCore,
) -> None:
    dataset_ids = [1, 2, 3]
    team_slug = "test-team"
    filters = {
        "not_statuses": ["uploading", "annotate"],
        "not_assignees": [123, 456, 789],
        "item_ids": [
            ("00000000-0000-0000-0000-000000000000"),
            ("00000000-0000-0000-0000-000000000000"),
        ],
    }
    path = "/test/path"

    responses.add(
        responses.POST,
        base_client.config.api_endpoint + "v2/teams/test-team/items/path",
        json={"affected_item_count": 2},
        status=200,
    )

    response = move_list_of_items_to_folder(
        client=base_client,
        team_slug=team_slug,
        dataset_ids=dataset_ids,
        filters=filters,
        path=path,
    )

    assert response == {"affected_item_count": 2}


@responses.activate
def test_move_list_of_items_to_folder_raises_on_incorrect_parameters(
    base_client: ClientCore,
) -> None:
    dataset_ids = [1, 2, 3]
    team_slug = "test-team"
    path = "/test/path"

    with pytest.raises(AssertionError):
        move_list_of_items_to_folder(
            client=base_client,
            team_slug=team_slug,
            dataset_ids=dataset_ids,
            path=path,
        )


@responses.activate
def test_move_list_of_items_to_folders_with_error_response(
    base_client: ClientCore,
) -> None:
    dataset_ids = [1, 2, 3]
    team_slug = "test-team"
    filters = {
        "not_statuses": ["uploading", "annotate"],
        "not_assignees": [123, 456, 789],
        "item_ids": [
            ("00000000-0000-0000-0000-000000000000"),
            ("00000000-0000-0000-0000-000000000000"),
        ],
    }
    path = "/test/path"

    responses.add(
        responses.POST,
        base_client.config.api_endpoint + "v2/teams/test-team/items/path",
        json={"error": "Bad Request"},
        status=400,
    )

    with pytest.raises(BadRequest):
        move_list_of_items_to_folder(
            client=base_client,
            team_slug=team_slug,
            dataset_ids=dataset_ids,
            path=path,
            filters=filters,
        )



from typing import Dict

import pytest
import responses

from darwin.future.core.client import ClientCore
from darwin.future.core.items.delete_items import delete_list_of_items
from darwin.future.data_objects.typing import UnknownType
from darwin.future.exceptions import BadRequest
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_delete_items_including_filters(base_client: ClientCore) -> None:
    dataset_ids = [1, 2, 3]
    team_slug = "test-team"
    filters: Dict[str, UnknownType] = {
        "not_statuses": ["uploading", "annotate"],
        "not_assignees": [123, 456, 789],
        "item_ids": [
            ("00000000-0000-0000-0000-000000000000"),
            ("00000000-0000-0000-0000-000000000000"),
        ],
    }

    responses.add(
        responses.DELETE,
        base_client.config.api_endpoint + "v2/teams/test-team/items",
        json={"affected_item_count": 2},
        status=200,
    )

    response = delete_list_of_items(
        client=base_client,
        team_slug=team_slug,
        dataset_ids=dataset_ids,
        filters=filters,
    )

    assert response == {"affected_item_count": 2}


@responses.activate
def test_delete_items_raises_on_incorrect_parameters(
    base_client: ClientCore,
) -> None:
    dataset_ids = [1, 2, 3]
    team_slug = "test-team"
    filters: Dict[str, UnknownType] = {}

    with pytest.raises(AssertionError):
        delete_list_of_items(
            client=base_client,
            team_slug=team_slug,
            dataset_ids=dataset_ids,
            filters=filters,
        )


@responses.activate
def test_delete_items_with_error_response(base_client: ClientCore) -> None:
    dataset_ids = [1, 2, 3]
    team_slug = "test-team"
    filters: Dict[str, UnknownType] = {
        "not_statuses": ["uploading", "annotate"],
        "not_assignees": [123, 456, 789],
        "item_ids": [
            ("00000000-0000-0000-0000-000000000000"),
            ("00000000-0000-0000-0000-000000000000"),
        ],
    }

    responses.add(
        responses.DELETE,
        base_client.config.api_endpoint + "v2/teams/test-team/items",
        json={"error": "Bad Request"},
        status=400,
    )

    with pytest.raises(BadRequest):
        delete_list_of_items(
            client=base_client,
            team_slug=team_slug,
            dataset_ids=dataset_ids,
            filters=filters,
        )



# @see: GraphotateWeb.Schemas.DatasetsV2.ItemRegistration.ExistingItem
from typing import Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import ValidationInfo, field_validator, model_validator

from darwin.datatypes import NumberLike
from darwin.future.data_objects.pydantic_base import DefaultDarwin
from darwin.future.data_objects.typing import UnknownType

ItemFrameRate = Union[NumberLike, Literal["native"]]


def validate_no_slashes(v: UnknownType) -> str:
    assert isinstance(v, str), "Must be a string"
    assert len(v) > 0, "cannot be empty"
    assert "/" not in v, "cannot contain slashes"

    return v


class ItemLayout(DefaultDarwin):
    # GraphotateWeb.Schemas.DatasetsV2.Common.ItemLayoutV1

    # Required fields
    slots: List[str]
    type: Literal["grid", "horizontal", "vertical", "simple"]
    version: Literal[1, 2]

    # Required only in version 2
    layout_shape: Optional[List[int]] = None

    # TODO[pydantic]: We couldn't refactor the `validator`, please replace it by `field_validator` manually.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-validators for more information.
    @field_validator("layout_shape")
    def layout_validator(cls, value: Dict, values: ValidationInfo) -> Dict:
        if not value and values.data.get("version") == 2:
            raise ValueError("layout_shape must be specified for version 2 layouts")

        return value


class ItemSlot(DefaultDarwin):
    # GraphotateWeb.Schemas.DatasetsV2.ItemRegistration.ExistingSlot

    # Required fields
    slot_name: str
    file_name: str
    fps: Optional[ItemFrameRate] = None

    # Optional fields
    storage_key: Optional[str] = None
    as_frames: Optional[bool] = None
    extract_views: Optional[bool] = None
    metadata: Optional[Dict[str, UnknownType]] = None
    tags: Optional[Union[List[str], Dict[str, str]]] = None
    type: Optional[Literal["image", "video", "pdf", "dicom"]] = None

    @field_validator("slot_name")
    @classmethod
    def validate_slot_name(cls, v: UnknownType) -> str:
        assert isinstance(v, str), "slot_name must be a string"
        assert len(v) > 0, "slot_name cannot be empty"
        return v

    @classmethod
    def validate_fps(cls, values: dict) -> dict:
        value = values.get("fps")

        if value is None:
            values["fps"] = 0
            return values

        assert isinstance(value, (int, float, str)), "fps must be a number or 'native'"
        if isinstance(value, str):
            assert value == "native", "fps must be 'native' or a number greater than 0"
        elif isinstance(value, (int, float)):
            type = values.get("type")
            if type == "image":
                assert value == 0 or value == 1.0, "fps must be '0' or '1.0' for images"
            else:
                assert value >= 0, "fps must be greater than or equal to 0 for videos"

        return values

    @classmethod
    def infer_type(cls, values: Dict[str, UnknownType]) -> Dict[str, UnknownType]:
        file_name = values.get("file_name")

        if file_name is not None:
            # TODO - Review types
            if file_name.endswith((".jpg", ".jpeg", ".png", ".bmp", ".gif")):
                values["type"] = "image"
            elif file_name.endswith(".pdf"):
                values["type"] = "pdf"
            elif file_name.endswith((".dcm", ".nii", ".nii.gz")):
                values["type"] = "dicom"
            elif file_name.endswith((".mp4", ".avi", ".mov", ".wmv", ".mkv")):
                values["type"] = "video"

        return values

    @model_validator(mode="before")
    def root(cls, values: Dict) -> Dict:
        values = cls.infer_type(values)
        values = cls.validate_fps(values)

        return values


class UploadItem(DefaultDarwin):
    # GraphotateWeb.Schemas.DatasetsV2.ItemRegistration.NewItem

    # Required fields
    name: str
    slots: List[ItemSlot] = []

    # Optional fields
    description: Optional[str] = None
    path: str = "/"
    tags: Optional[Union[List[str], Dict[str, str]]] = []
    layout: Optional[ItemLayout] = None

    @field_validator("name")
    @classmethod
    def validate_name(cls, v: UnknownType) -> str:
        return validate_no_slashes(v)


class ItemCore(DefaultDarwin):
    # GraphotateWeb.Schemas.DatasetsV2.ItemRegistration.NewItem

    # Required fields
    name: str
    id: UUID
    slots: List[ItemSlot] = []
    path: str = "/"
    dataset_id: int
    processing_status: str

    # Optional fields
    archived: Optional[bool] = False
    priority: Optional[int] = None
    tags: Optional[Union[List[str], Dict[str, str]]] = []
    layout: Optional[ItemLayout] = None

    @field_validator("name")
    @classmethod
    def validate_name(cls, v: UnknownType) -> str:
        return validate_no_slashes(v)


class Folder(DefaultDarwin):
    dataset_id: int
    filtered_item_count: int
    path: str
    unfiltered_item_count: int



from __future__ import annotations

from typing import List, Optional

from pydantic import field_validator

from darwin.future.core.client import ClientCore
from darwin.future.core.team.get_raw import get_team_raw
from darwin.future.data_objects.dataset import DatasetList
from darwin.future.data_objects.team_member_role import TeamMemberRole
from darwin.future.data_objects.validators import parse_name
from darwin.future.pydantic_base import DefaultDarwin


class TeamMemberCore(DefaultDarwin):
    """A class to manage all the information around a team member on the darwin platform
    Attributes
    ----------
    name: str
    """

    email: str
    first_name: str
    last_name: str
    role: TeamMemberRole
    team_id: int
    id: int
    user_id: int


TeamMemberList = List[TeamMemberCore]


class TeamCore(DefaultDarwin):
    """
    A class to manage all the information around a Team on the darwin platform
    including validation

    Attributes
    ----------
    slug : str
    datasets: Optional[List[Dataset]] = None
        - a list of datasets linked to the team
    members: Optional[List[Release]] = None
        - a list of members linked to a team
    Methods
    ----------
    _slug_validator: validates and auto formats the slug variable
    """

    name: str
    slug: str
    id: int
    datasets: Optional[DatasetList] = None
    members: Optional[List[TeamMemberCore]] = None
    default_role: TeamMemberRole = TeamMemberRole.USER

    # Data Validation
    _slug_validator = field_validator("slug")(parse_name)

    @staticmethod
    def from_client(client: ClientCore, team_slug: Optional[str] = None) -> TeamCore:
        """Returns the team with the given slug from the client

        Args:
            client (Client): Core client object
            team_slug (Optional[str], optional): team slug str, Defaults to None.

        Returns:
            Team: Team object retrieved from the client with the given slug
        """
        if not team_slug:
            team_slug = client.config.default_team
        url = client.config.api_endpoint + f"teams/{team_slug}"
        return TeamCore.model_validate(get_team_raw(client.session, url))



from typing import Optional, Union

from darwin.future.core.client import ClientCore
from darwin.future.core.types.common import JSONDict
from darwin.future.data_objects.properties import FullProperty, PropertyValue


def update_property(
    client: ClientCore,
    params: Union[FullProperty, JSONDict],
    team_slug: Optional[str] = None,
) -> FullProperty:
    """
    Updates a property for the specified team slug.

    Parameters:
        client (ClientCore): The client to use for the request.
        team_slug (Optional[str]): The slug of the team to get. If not specified, the
            default team from the client's config will be used.
        params (Optional[JSONType]): The JSON data to use for the request.

    Returns:
        FullProperty: FullProperty object for the created property.

    Raises:
        HTTPError: If the response status code is not in the 200-299 range.
    """
    if not team_slug:
        team_slug = client.config.default_team
    if isinstance(params, FullProperty):
        id, params = params.to_update_endpoint()
    else:
        id = params.get("id")
        del params["id"]
    response = client.put(f"/v2/teams/{team_slug}/properties/{id}", data=params)
    assert isinstance(response, dict)
    return FullProperty.model_validate(response)


def update_property_value(
    client: ClientCore,
    params: Union[PropertyValue, JSONDict],
    item_id: str,
    team_slug: Optional[str] = None,
) -> PropertyValue:
    """
    Updates a property value for the specified property id.

    Parameters:
        client (ClientCore): The client to use for the request.
        team_slug (Optional[str]): The slug of the team to get. If not specified, the
            default team from the client's config will be used.
        params (Optional[JSONType]): The JSON data to use for the request.

    Returns:
        FullProperty: FullProperty object for the created property.

    Raises:
        HTTPError: If the response status code is not in the 200-299 range.
    """
    if not team_slug:
        team_slug = client.config.default_team
    if isinstance(params, PropertyValue):
        id, params = params.to_update_endpoint()
    else:
        id = params.get("id")
        del params["id"]
    response = client.put(
        f"/v2/teams/{team_slug}/properties/{item_id}/property_values/{id}", data=params
    )
    assert isinstance(response, dict)
    return PropertyValue.model_validate(response)



import random
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Tuple, Union

import numpy as np
import torch
import torchvision.transforms as transforms
import torchvision.transforms.functional as F
from PIL import Image as PILImage

from darwin.torch.utils import convert_segmentation_to_mask, flatten_masks_by_category

# Optional dependency
try:
    import albumentations as A
    from albumentations import Compose
except ImportError:
    A = None

from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from albumentations.pytorch import ToTensorV2

    AType = Type[ToTensorV2]
else:
    AType = Type[None]
    Compose = Type[None]

TargetKey = Literal["boxes", "labels", "mask", "masks", "image_id", "area", "iscrowd"]


TargetType = Dict[TargetKey, torch.Tensor]


class Compose(transforms.Compose):
    """
    Composes a sequence of Transformations.
    """

    def __call__(self, image: PILImage.Image, target: Optional[TargetType] = None):
        if target is None:
            return super(Compose, self).__call__(image)
        for transform in self.transforms:
            image, target = transform(image, target)
        return image, target


class RandomHorizontalFlip(transforms.RandomHorizontalFlip):
    """
    Allows for horizontal flipping of an image, randomly.
    """

    def forward(
        self, image: torch.Tensor, target: Optional[TargetType] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]:
        """
        May or may not horizontally flip an image depending on a random factor.

        Parameters
        ----------
        image : torch.Tensor
            Image ``Tensor`` to flip.
        target : Optional[TargetType] = None
            The target.

        Returns
        -------
        Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]
            Will return a single image ``Tensor`` if the flip did not happen, or a tuple of the
            image tensor and the target type if the flip did happen.

        """
        if random.random() < self.p:
            image = F.hflip(image)
            if target is None:
                return image

            if "boxes" in target:
                bbox = target["boxes"]
                bbox[:, [0, 2]] = image.size[0] - bbox[:, [2, 0]]
                target["boxes"] = bbox
            for k in ["mask", "masks"]:
                if k in target:
                    target[k] = target[k].flip(-1)
            return image, target

        if target is None:
            return image

        return image, target


class RandomVerticalFlip(transforms.RandomVerticalFlip):
    """
    Allows for vertical flipping of an image, randomly.
    """

    def forward(
        self, image: torch.Tensor, target: Optional[TargetType] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]:
        """
        May or may not vertically flip an image depending on a random factor.

        Parameters
        ----------
        image : torch.Tensor
            Image ``Tensor`` to flip.
        target : Optional[TargetType] = None
            The target.

        Returns
        -------
        Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]
            Will return a single image ``Tensor`` if the flip did not happen, or a tuple of the
            image tensor and the target type if the flip did happen.

        """
        if random.random() < self.p:
            image = F.vflip(image)
            if target is None:
                return image

            if "boxes" in target:
                bbox = target["boxes"]
                bbox[:, [1, 3]] = image.size[1] - bbox[:, [1, 3]]
                target["boxes"] = bbox
            for k in ["mask", "masks"]:
                if k in target:
                    target[k] = target[k].flip(-2)
            return image, target

        if target is None:
            return image
        return image, target


class ColorJitter(transforms.ColorJitter):
    """
    Jitters the colors of the given transformation.
    """

    def __call__(
        self, image: PILImage.Image, target: Optional[TargetType] = None
    ) -> Union[PILImage.Image, Tuple[PILImage.Image, TargetType]]:
        transform = self.get_params(
            self.brightness, self.contrast, self.saturation, self.hue
        )
        image = transform(image)
        if target is None:
            return image
        return image, target


class ToTensor(transforms.ToTensor):
    """
    Converts given ``PILImage`` to a ``Tensor``.
    """

    def __call__(
        self, image: PILImage.Image, target: Optional[TargetType] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]:
        image_tensor: torch.Tensor = F.to_tensor(image)
        if target is None:
            return image_tensor
        return image_tensor, target


class ToPILImage(transforms.ToPILImage):
    """
    Converts given ``Tensor`` to a ``PILImage``.
    """

    def __call__(
        self, image: torch.Tensor, target: Optional[TargetType] = None
    ) -> Union[PILImage.Image, Tuple[PILImage.Image, TargetType]]:
        pil_image: PILImage.Image = F.to_pil_image(image)
        if target is None:
            return pil_image
        return pil_image, target


class Normalize(transforms.Normalize):
    """
    Normalizes the given ``Tensor``.
    """

    def __call__(
        self, tensor: torch.Tensor, target: Optional[TargetType] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]:
        tensor = F.normalize(tensor, self.mean, self.std, self.inplace)

        if target is None:
            return tensor
        return tensor, target


class ConvertPolygonsToInstanceMasks(object):
    """
    Converts given polygon to an ``InstanceMask``.
    """

    def __call__(
        self, image: PILImage.Image, target: TargetType
    ) -> Tuple[PILImage.Image, TargetType]:
        w, h = image.size

        image_id = target["image_id"]
        image_id = torch.tensor([image_id])

        annotations = target.pop("annotations")

        annotations = [obj for obj in annotations if obj.get("iscrowd", 0) == 0]

        boxes = [obj["bbox"] for obj in annotations]
        # guard against no boxes via resizing
        boxes = torch.as_tensor(boxes, dtype=torch.float32).reshape(-1, 4)

        classes = [obj["category_id"] for obj in annotations]
        classes = torch.tensor(classes, dtype=torch.int64)

        segmentations = [obj["segmentation"] for obj in annotations]
        masks = convert_segmentation_to_mask(segmentations, h, w)

        keypoints = None
        if annotations and "keypoints" in annotations[0]:
            keypoints = [obj["keypoints"] for obj in annotations]
            keypoints = torch.as_tensor(keypoints, dtype=torch.float32)
            num_keypoints = keypoints.shape[0]
            if num_keypoints:
                keypoints = keypoints.view(num_keypoints, -1, 3)

        target["boxes"] = boxes
        target["labels"] = classes
        target["masks"] = masks
        target["image_id"] = image_id
        if keypoints is not None:
            target["keypoints"] = keypoints

        # Remove boxes with widht or height zero
        keep = (boxes[:, 3] > 0) & (boxes[:, 2] > 0)
        boxes = boxes[keep]
        classes = classes[keep]
        masks = masks[keep]
        if keypoints is not None:
            keypoints = keypoints[keep]

        # conversion to coco api
        area = torch.tensor([obj["area"] for obj in annotations])
        iscrowd = torch.tensor([obj.get("iscrowd", 0) for obj in annotations])
        target["area"] = area
        target["iscrowd"] = iscrowd

        return image, target


class ConvertPolygonsToSemanticMask(object):
    """
    Converts given polygon to an ``SemanticMask``.
    """

    def __call__(
        self, image: PILImage.Image, target: TargetType
    ) -> Tuple[PILImage.Image, TargetType]:
        w, h = image.size
        image_id = target["image_id"]
        image_id = torch.tensor([image_id])

        annotations = target.pop("annotations")
        segmentations = [obj["segmentation"] for obj in annotations]
        cats = [obj["category_id"] for obj in annotations]
        if segmentations:
            masks = convert_segmentation_to_mask(segmentations, h, w)
            # merge all instance masks into a single segmentation map
            # with its corresponding categories
            mask = flatten_masks_by_category(masks, cats)

        else:
            mask = torch.zeros((h, w), dtype=torch.uint8)

        target["mask"] = mask
        target["image_id"] = image_id
        return image, target


class ConvertPolygonToMask(object):
    """
    Converts given polygon to a ``Mask``.
    """

    def __call__(
        self, image: PILImage.Image, annotation: Dict[str, Any]
    ) -> Tuple[PILImage.Image, PILImage.Image]:
        w, h = image.size
        segmentations = [obj["segmentation"] for obj in annotation]
        cats = [obj["category_id"] for obj in annotation]
        if segmentations:
            masks = convert_segmentation_to_mask(segmentations, h, w)
            # merge all instance masks into a single segmentation map
            # with its corresponding categories
            target = flatten_masks_by_category(masks, cats)

        else:
            target = torch.zeros((h, w), dtype=torch.uint8)
        target = PILImage.fromarray(target.numpy())
        return image, target


class AlbumentationsTransform:
    """
    Wrapper class for Albumentations augmentations.
    """

    def __init__(self, transform: Compose):
        self._check_albumentaion_dependency()
        self.transform = transform

    @classmethod
    def from_path(cls, config_path: str) -> "AlbumentationsTransform":
        config_path = Path(config_path)
        try:
            transform = A.load(str(config_path))
            return cls(transform)
        except Exception as e:
            raise ValueError(f"Invalid config path: {config_path}. Error: {e}")

    @classmethod
    def from_dict(cls, alb_dict: dict) -> "AlbumentationsTransform":
        try:
            transform = A.from_dict(alb_dict)
            return cls(transform)
        except Exception as e:
            raise ValueError(f"Invalid albumentations dictionary. Error: {e}")

    def __call__(self, image, annotation: dict = None) -> tuple:
        np_image = np.array(image)
        if annotation is None:
            annotation_dict = {}
        else:
            annotation_dict = annotation

        albu_data = self._pre_process(np_image, annotation_dict)
        transformed_data = self.transform(**albu_data)
        image, transformed_annotation = self._post_process(
            transformed_data, annotation_dict
        )

        if annotation is None:
            return image

        return image, transformed_annotation

    def _pre_process(self, image: np.ndarray, annotation: dict) -> dict:
        """
        Prepare image and annotation for albumentations transformation.
        """
        albumentation_dict = {"image": image}

        boxes = annotation.get("boxes")
        if boxes is not None:
            albumentation_dict["bboxes"] = boxes.numpy().tolist()

        labels = annotation.get("labels")
        if labels is not None:
            albumentation_dict["labels"] = labels.tolist()

        masks = annotation.get("masks")
        if (
            masks is not None and masks.numel() > 0
        ):  # using numel() to check if tensor is non-empty
            if isinstance(masks, torch.Tensor):
                masks = masks.numpy()
            if masks.ndim == 3:  # Ensure masks is a list of numpy arrays
                masks = [masks[i] for i in range(masks.shape[0])]
            albumentation_dict["masks"] = masks

        return albumentation_dict

    def _post_process(self, albumentation_output: dict, annotation: dict) -> tuple:
        """
        Process the output of albumentations transformation back to desired format.
        """
        output_annotation = {}
        image = albumentation_output["image"]

        bboxes = albumentation_output.get("bboxes")
        if bboxes is not None:
            output_annotation["boxes"] = torch.tensor(bboxes)

        labels = albumentation_output.get("labels")
        if labels is not None:
            output_annotation["labels"] = torch.tensor(labels)

        masks = albumentation_output.get("masks")
        if masks is not None:
            if isinstance(masks[0], np.ndarray):
                output_annotation["masks"] = torch.tensor(np.array(masks))
            else:
                output_annotation["masks"] = torch.stack(masks)
        elif "masks" in annotation:
            output_annotation["masks"] = torch.tensor([])

        if "area" in annotation:
            if "masks" in output_annotation and output_annotation["masks"].numel() > 0:
                output_annotation["area"] = torch.sum(
                    output_annotation["masks"], dim=[1, 2]
                )
            elif "boxes" in output_annotation and len(output_annotation["boxes"]) > 0:
                output_annotation["area"] = (
                    output_annotation["boxes"][:, 2] * output_annotation["boxes"][:, 3]
                )
            else:
                output_annotation["area"] = torch.tensor([])

        # Copy other metadata from original annotation
        for key, value in annotation.items():
            output_annotation.setdefault(key, value)

        return image, output_annotation

    def _check_albumentaion_dependency(self):
        if A is None:
            raise ImportError(
                "The albumentations library is not installed. "
                "To use this function, install it with pip install albumentations, "
                "or install the ml extras of this package."
            )



import responses
from responses.matchers import query_param_matcher

from darwin.future.core.client import ClientCore
from darwin.future.core.properties import (
    get_property_by_id,
    get_team_full_properties,
    get_team_properties,
)
from darwin.future.core.types.common import QueryString
from darwin.future.data_objects.properties import FullProperty
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_get_team_properties(
    base_client: ClientCore, base_property_object: FullProperty
) -> None:
    # Mocking the response using responses library
    base_property_object.options = None
    base_property_object.property_values = None
    response_data = {"properties": [base_property_object.model_dump(mode="json")]}
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/properties",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    properties = get_team_properties(base_client)

    # Assertions
    assert isinstance(properties, list)
    assert all(isinstance(property, FullProperty) for property in properties)
    assert properties[0] == base_property_object


@responses.activate
def test_get_team_full_properties(
    base_client: ClientCore, base_property_object: FullProperty
) -> None:
    # Mocking the response using responses library
    response_data = {"properties": [base_property_object.model_dump(mode="json")]}
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/properties",
        match=[
            query_param_matcher({"include_values": "true"}),
        ],
        json=response_data,
        status=200,
    )
    params = QueryString({"include_values": True})
    # Call the function being tested
    properties = get_team_full_properties(base_client, params=params)

    # Assertions
    assert isinstance(properties, list)
    assert all(isinstance(property, FullProperty) for property in properties)
    assert properties[0] == base_property_object


@responses.activate
def test_get_property_by_id(
    base_client: ClientCore, base_property_object: FullProperty
) -> None:
    # Mocking the response using responses library
    property_id = "0"
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/properties/{property_id}",
        json=base_property_object.model_dump(mode="json"),
        status=200,
    )

    # Call the function being tested
    property = get_property_by_id(base_client, property_id)

    # Assertions
    assert isinstance(property, FullProperty)
    assert property == base_property_object



from __future__ import annotations

from math import floor

from pydantic import NonNegativeInt, PositiveInt

from darwin.future.core.types.common import QueryString
from darwin.future.data_objects.pydantic_base import DefaultDarwin


def must_be_positive(v: int) -> int:
    if v is not None and v < 0:
        raise ValueError("Value must be positive")
    return v


class Page(DefaultDarwin):
    offset: NonNegativeInt = 0
    size: PositiveInt = 500

    def get_required_page(self, item_index: int) -> Page:
        """
        Get the page that contains the item at the specified index

        Args:
            item_index (int): The index of the item

        Returns:
            Page: The page that contains the item
        """
        assert self.size is not None
        required_offset = floor(item_index / self.size) * self.size
        return Page(offset=required_offset, size=self.size)

    def to_query_string(self) -> QueryString:
        """
        Generate a query string from the page object, some fields are not included if they are None,
        and certain fields are renamed. Outgoing and incoming query strings are different and require
        dropping certain fields

        Returns:
            QueryString: Outgoing query string
        """
        qs_dict = {"page[offset]": str(self.offset), "page[size]": str(self.size)}
        return QueryString(qs_dict)

    def increment(self) -> None:
        """
        Increment the page offset by the page size
        """
        self.offset += self.size



from datetime import datetime, timezone
from typing import Dict, List
from uuid import UUID

from darwin.exceptions import DarwinException
from darwin.future.core.types.query import Query, QueryFilter
from darwin.future.core.workflows import list_workflows
from darwin.future.data_objects.workflow import WFStageCore
from darwin.future.helpers.exception_handler import handle_exception
from darwin.future.meta.objects.workflow import Workflow


class WorkflowQuery(Query[Workflow]):
    """
    WorkflowQuery object with methods to manage filters, retrieve data, and execute
    filters

    Methods
    -------

    collect: Executes the query and returns the filtered data
    """

    def _collect(self) -> Dict[int, Workflow]:
        workflows_core, exceptions = list_workflows(self.client)
        if exceptions:
            handle_exception(exceptions)
            raise DarwinException from exceptions[0]
        workflows = [
            Workflow(client=self.client, element=workflow, meta_params=self.meta_params)
            for workflow in workflows_core
        ]
        if not self.filters:
            self.filters = []

        for filter in self.filters:
            workflows = self._execute_filters(workflows, filter)

        return dict(enumerate(workflows))

    def _execute_filters(
        self, workflows: List[Workflow], filter: QueryFilter
    ) -> List[Workflow]:
        if filter.name == "id":
            id_to_find = UUID(filter.param)
            return [w for w in workflows if w.id == id_to_find]

        if filter.name == "inserted_at_start":
            start_date = datetime.fromisoformat(filter.param)
            return [
                w
                for w in workflows
                if w._element is not None
                and self._date_compare(w._element.inserted_at, start_date)
            ]

        if filter.name == "inserted_at_end":
            end_date = datetime.fromisoformat(filter.param)
            return [
                w
                for w in workflows
                if w._element is not None
                and self._date_compare(end_date, w._element.inserted_at)
            ]

        if filter.name == "updated_at_start":
            start_date = datetime.fromisoformat(filter.param)
            return [
                w
                for w in workflows
                if w._element is not None
                and self._date_compare(w._element.updated_at, start_date)
            ]

        if filter.name == "updated_at_end":
            end_date = datetime.fromisoformat(filter.param)
            return [
                w
                for w in workflows
                if w._element is not None
                and self._date_compare(end_date, w._element.updated_at)
            ]

        if filter.name == "dataset_id":
            datasets_to_find_id: List[int] = [int(s) for s in filter.param.split(",")]
            return [
                w
                for w in workflows
                if w._element is not None
                and w._element.dataset is not None
                and int(w._element.dataset.id) in datasets_to_find_id
            ]

        if filter.name == "dataset_name":
            datasets_to_find_name: List[str] = [str(s) for s in filter.param.split(",")]
            return [
                w
                for w in workflows
                if w._element is not None
                and str(w._element.dataset) in datasets_to_find_name
            ]

        if filter.name == "has_stages":
            stages_to_find = list(filter.param.split(","))
            return [
                w
                for w in workflows
                if w._element is not None
                and self._stages_contains(w._element.stages, stages_to_find)
            ]

        return self._generic_execute_filter(workflows, filter)

    @classmethod
    def _date_compare(cls, date1: datetime, date2: datetime) -> bool:
        return date1.astimezone(timezone.utc) >= date2.astimezone(timezone.utc)

    @classmethod
    def _stages_contains(
        cls, stages: List[WFStageCore], stages_to_find: List[str]
    ) -> bool:
        stage_ids = [str(s.id) for s in stages]
        return any(stage_to_find in stage_ids for stage_to_find in stages_to_find)



from typing import List
from uuid import UUID, uuid4

import responses
from pydantic import ValidationError

from darwin.future.core.client import ClientCore
from darwin.future.core.items import get_item_ids, get_item_ids_stage
from darwin.future.core.items.get import get_item, list_folders, list_items
from darwin.future.core.types.common import QueryString
from darwin.future.data_objects.item import Folder, ItemCore
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.core.items.fixtures import *


def test_get_item_ids(
    UUIDs: List[UUID], UUIDs_str: List[str], base_client: ClientCore
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint + "v2/teams/default-team/items/list_ids"
            "?dataset_ids=1337",
            json={"item_ids": UUIDs_str},
            status=200,
        )
        item_ids = get_item_ids(base_client, "default-team", "1337")
        assert item_ids == UUIDs


def test_get_item_ids_stage(
    UUIDs: List[UUID], UUIDs_str: List[str], base_client: ClientCore
) -> None:
    stage_id = str(uuid4())
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint + "v2/teams/default-team/items/ids"
            f"?workflow_stage_ids={stage_id}&dataset_ids=1337",
            json={"item_ids": UUIDs_str},
            status=200,
        )
        item_ids = get_item_ids_stage(base_client, "default-team", "1337", stage_id)
        assert item_ids == UUIDs


def test_get_item(
    base_items_json: List[dict], base_items: List[ItemCore], base_client: ClientCore
) -> None:
    uuid = str(base_items[0].id)
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint + f"v2/teams/default-team/items/{uuid}",
            json=base_items_json[0],
            status=200,
        )
        item = get_item(base_client, "default-team", uuid)
        assert item == base_items[0]


def test_list_items(
    base_items_json: List[dict], base_items: List[ItemCore], base_client: ClientCore
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint
            + "v2/teams/default-team/items?dataset_ids=1337",
            json={"items": base_items_json},
            status=200,
        )
        items, _ = list_items(base_client, "default-team", dataset_ids=[1337])
        for item, comparator in zip(items, base_items):
            assert item == comparator


def test_list_items_breaks(
    base_items_json: List[dict], base_client: ClientCore
) -> None:
    malformed = base_items_json.copy()
    del malformed[0]["name"]
    del malformed[0]["dataset_id"]
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint
            + "v2/teams/default-team/items?dataset_ids=1337",
            json={"items": base_items_json},
            status=200,
        )
        items, exceptions = list_items(base_client, "default-team", dataset_ids=[1337])
        assert len(exceptions) == 1
        assert isinstance(exceptions[0], ValidationError)


def test_list_folders(
    base_folders_json: List[dict], base_folders: List[Folder], base_client: ClientCore
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint
            + "v2/teams/default-team/items/folders?dataset_ids=1337",
            json={"folders": base_folders_json},
            status=200,
        )
        folders, _ = list_folders(
            base_client, "default-team", QueryString({"dataset_ids": "1337"})
        )
        for folder, comparator in zip(folders, base_folders):
            assert folder == comparator


def test_list_folders_breaks(
    base_folders_json: List[dict], base_client: ClientCore
) -> None:
    malformed = base_folders_json.copy()
    del malformed[0]["dataset_id"]
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_client.config.api_endpoint
            + "v2/teams/default-team/items/folders?dataset_ids=1337",
            json={"folders": malformed},
            status=200,
        )
        folders, exceptions = list_folders(
            base_client, "default-team", QueryString({"dataset_ids": "1337"})
        )
        assert len(exceptions) == 1
        assert isinstance(exceptions[0], ValidationError)



from pathlib import Path
from typing import Callable, Dict, Iterable, List

from darwin.datatypes import AnnotationFile

ClassIndex = Dict[str, int]


def build_class_index(
    annotation_files: Iterable[AnnotationFile],
    include_types: List[str] = ["bounding_box", "polygon"],
) -> ClassIndex:
    classes = set()
    for annotation_file in annotation_files:
        for annotation in annotation_file.annotations:
            if annotation.annotation_class.annotation_type in include_types:
                classes.add(annotation.annotation_class.name)
    return {k: v for (v, k) in enumerate(sorted(classes))}


def export_file(
    annotation_file: AnnotationFile,
    class_index: ClassIndex,
    output_dir: Path,
    build_function: Callable[[AnnotationFile, ClassIndex], str],
) -> None:
    txt = build_function(annotation_file, class_index)

    # Just using `.with_suffix(".txt")` would remove all suffixes, so we need to
    # do it manually.

    filename = annotation_file.path.name
    filename_to_write = (
        filename.replace(".json", ".txt") if ".json" in filename else filename + ".txt"
    )
    output_file_path = output_dir / filename_to_write

    output_file_path.parent.mkdir(parents=True, exist_ok=True)
    with open(output_file_path, "w") as f:
        f.write(txt)


def save_class_index(class_index: ClassIndex, output_dir: Path) -> None:
    sorted_items = sorted(class_index.items(), key=lambda item: item[1])

    with open(output_dir / "darknet.labels", "w") as f:
        for class_name, _ in sorted_items:
            f.write(f"{class_name}\n")



from __future__ import annotations

from typing import List, Literal, Optional, Union

from pydantic import BaseModel, Field, ValidationInfo, field_validator

from darwin.future.data_objects.properties import SelectedProperty


class Point(BaseModel):
    x: float
    y: float

    def __add__(self, other: Point) -> Point:
        return Point(x=self.x + other.x, y=self.y + other.y)

    def __sub__(self, other: Point) -> Point:
        return Point(x=self.x - other.x, y=self.y - other.y)


PolygonPath = List[Point]


class Polygon(BaseModel):
    paths: List[PolygonPath]

    def bounding_box(self) -> BoundingBox:
        h, w, x, y = 0.0, 0.0, 0.0, 0.0
        for polygon_path in self.paths:
            for point in polygon_path:
                h = max(h, point.y)
                w = max(w, point.x)
                x = min(x, point.x)
                y = min(y, point.y)
        return BoundingBox(h=h, w=w, x=x, y=y)

    @property
    def is_complex(self) -> bool:
        return len(self.paths) > 1

    @property
    def center(self) -> Point:
        return self.bounding_box().center


class AnnotationBase(BaseModel):
    id: str
    name: str
    properties: Optional[SelectedProperty] = None
    slot_names: Optional[List[str]] = None

    @field_validator("id")
    def validate_id_is_UUID(cls, v: str) -> str:
        assert len(v) == 36
        assert "-" in v
        return v


class BoundingBox(BaseModel):
    h: float
    w: float
    x: float
    y: float

    @property
    def center(self) -> Point:
        return Point(x=self.x + self.w / 2, y=self.y + self.h / 2)


class BoundingBoxAnnotation(AnnotationBase):
    bounding_box: BoundingBox


class Ellipse(BaseModel):
    center: Point
    radius: Point
    angle: float


class EllipseAnnotation(AnnotationBase):
    ellipse: Ellipse


class PolygonAnnotation(AnnotationBase):
    polygon: Polygon
    bounding_box: Optional[BoundingBox] = Field(default=None, validate_default=True)

    @field_validator("bounding_box")
    def validate_bounding_box(
        cls, v: Optional[BoundingBox], values: ValidationInfo
    ) -> BoundingBox:
        if v is None:
            assert "polygon" in values.data
            assert isinstance(values.data["polygon"], Polygon)
            v = values.data["polygon"].bounding_box()
        return v


class FrameAnnotation(AnnotationBase):
    frames: List
    interpolated: bool
    interpolate_algorithm: str
    ranges: List[int]


AllowedAnnotation = Union[
    PolygonAnnotation, BoundingBoxAnnotation, EllipseAnnotation, FrameAnnotation
]


class Item(BaseModel):
    name: str
    path: str


class DarwinV2(BaseModel):
    version: Literal["2.0"] = "2.0"
    schema_ref: str
    item: dict
    annotations: List[AllowedAnnotation]

    @field_validator("schema_ref")
    def validate_schema_ref(cls, v: str) -> str:
        assert v.startswith("http")
        return v



from typing import Generator
from unittest.mock import Mock, patch

from pytest import fixture, raises
from responses import RequestsMock

from darwin.future.core.client import ClientCore, DarwinConfig
from darwin.future.data_objects.dataset import DatasetCore
from darwin.future.data_objects.team import TeamMemberCore
from darwin.future.meta.client import Client
from darwin.future.meta.objects.dataset import Dataset
from darwin.future.meta.objects.team import Team
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.meta.objects.fixtures import *


@fixture
def _delete_by_slug_mock() -> Generator:
    with patch.object(Team, "_delete_dataset_by_slug") as mock:
        yield mock


@fixture
def _delete_by_id_mock() -> Generator:
    with patch.object(Team, "_delete_dataset_by_id") as mock:
        yield mock


def test_team_meta_collects_members(
    base_meta_team: Team,
    base_client: ClientCore,
    base_team_member: TeamMemberCore,
    base_team_member_json: dict,
) -> None:
    with RequestsMock() as rsps:
        endpoint = base_client.config.api_endpoint + "memberships"
        rsps.add(rsps.GET, endpoint, json=[base_team_member_json])
        members = base_meta_team.members._collect()
        assert len(members) == 1
        assert members[0]._element == base_team_member


# `delete_dataset` tests
def test_delete_dataset_returns_exceptions_thrown(
    base_config: DarwinConfig, _delete_by_id_mock: Mock, _delete_by_slug_mock: Mock
) -> None:
    _delete_by_slug_mock.side_effect = Exception("test exception")

    valid_client = Client(base_config)
    with raises(Exception):
        _ = Team.delete_dataset(valid_client, "test_dataset")

    assert _delete_by_slug_mock.call_count == 1
    assert _delete_by_id_mock.call_count == 0


def test_delete_dataset_calls_delete_by_slug_as_appropriate(
    base_config: DarwinConfig, _delete_by_id_mock: Mock, _delete_by_slug_mock: Mock
) -> None:
    valid_client = Client(base_config)

    _ = Team.delete_dataset(valid_client, "test_dataset")

    assert _delete_by_slug_mock.call_count == 1
    assert _delete_by_id_mock.call_count == 0


def test_delete_dataset_calls_delete_by_id_as_appropriate(
    base_config: DarwinConfig, _delete_by_id_mock: Mock, _delete_by_slug_mock: Mock
) -> None:
    valid_client = Client(base_config)

    _ = Team.delete_dataset(valid_client, 1)

    assert _delete_by_slug_mock.call_count == 0
    assert _delete_by_id_mock.call_count == 1


def test_delete_by_slug__returns_dataset_deleted_if_dataset_found(
    base_config: DarwinConfig,
) -> None:
    valid_client = Client(base_config)
    valid_slug = "test_dataset"

    base_url = base_config.base_url + "api/datasets"

    with RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_url + "?id=test_dataset",
            json={"id": 1, "name": "Test Dataset", "slug": "test_dataset"},
            status=200,
        )
        rsps.add(
            rsps.PUT,
            base_url + "/1/archive",
            json={"id": 1, "name": "Test Dataset", "slug": "test_dataset"},
            status=200,
        )
        dataset_deleted = Team._delete_dataset_by_slug(valid_client, valid_slug)

        assert dataset_deleted == 1


def test_delete_by_id_returns_dataset_deleted_if_dataset_found(
    base_config: DarwinConfig,
) -> None:
    valid_client = Client(base_config)
    valid_id = 1

    base_url = base_config.base_url + "api/datasets"

    with RequestsMock() as rsps:
        rsps.add(
            rsps.PUT,
            base_url + "/1/archive",
            json={"id": 1, "name": "Test Dataset", "slug": "test_dataset"},
            status=200,
        )
        dataset_deleted = Team._delete_dataset_by_id(valid_client, valid_id)

        assert dataset_deleted == 1


# Test `_delete_by_id`
def test_delete_by_id_raises_exception_if_not_passed_int_and_client(
    base_config: DarwinConfig,
) -> None:
    valid_client = Client(base_config)
    valid_id = 1
    invalid_client = "client"
    invalid_id = "1"

    with raises(AssertionError):
        Team._delete_dataset_by_id(valid_client, invalid_id)  # type: ignore

    with raises(AssertionError):
        Team._delete_dataset_by_id(invalid_client, valid_id)  # type: ignore


# Test `_delete_by_slug`
def test_delete_by_slug_raises_exception_if_not_passed_str_and_client(
    base_config: DarwinConfig,
) -> None:
    valid_client = Client(base_config)
    valid_slug = "test_dataset"
    invalid_client = "client"
    invalid_slug = 1

    with raises(AssertionError):
        Team._delete_dataset_by_slug(valid_client, invalid_slug)  # type: ignore

    with raises(AssertionError):
        Team._delete_dataset_by_slug(invalid_client, valid_slug)  # type: ignore


def test_create_dataset(base_meta_team: Team, base_config: DarwinConfig) -> None:
    base_url = base_config.base_url + "api/datasets"
    valid_slug = "test_dataset"
    valid_name = "test dataset"
    with RequestsMock() as rsps:
        rsps.add(
            rsps.POST,
            base_url,
            json={"id": 1, "name": valid_name, "slug": valid_slug},
            status=201,
        )

        dataset_created = base_meta_team.create_dataset(valid_slug)
        assert dataset_created is not None
        assert isinstance(dataset_created, Dataset)
        assert isinstance(dataset_created._element, DatasetCore)
        assert dataset_created.id == 1
        assert dataset_created.name == valid_name
        assert dataset_created.slug == valid_slug


def test_team_str_method(base_meta_team: Team) -> None:
    assert (
        str(base_meta_team)
        == "Team\n\
- Team Name: test-team\n\
- Team Slug: test-team\n\
- Team ID: 0\n\
- 0 member(s)"
    )


def test_team_repr_method(base_meta_team: Team) -> None:
    assert repr(base_meta_team) == str(base_meta_team)



from __future__ import annotations

from pathlib import Path
from typing import Optional

from requests.adapters import Retry

from darwin.future.core.client import ClientCore, DarwinConfig
from darwin.future.meta.objects.team import Team


class Client(ClientCore):
    """
    The Darwin Client object. Provides access to Darwin's API.

    Args:
        ClientCore (Client): Generic ClientCore object expanded by DarwinConfig object
            return type

    Returns:
        _type_: Client

    Attributes:
        _team (Optional[Team]): The team associated with the client.

    Methods:
        local(cls) -> Client: Creates a new client object with a local DarwinConfig.
        from_api_key(cls, api_key: str, datasets_dir: Optional[Path] = None) -> Client:
            Creates a new client object with a DarwinConfig from an API key.

    Example Usage:
        # Create a new client object with a local DarwinConfig
        client = Client.local()

        # Create a new client object with a DarwinConfig from an API key
        client = Client.from_api_key(api_key="my_api_key", datasets_dir="path/to/datasets/dir")

        # Access the team via chaining
        team = client.team # returns a Team object which can be chained further
    """

    def __init__(self, config: DarwinConfig, retries: Optional[Retry] = None) -> None:
        self._team: Optional[Team] = None
        super().__init__(config, retries=retries)

    @classmethod
    def local(cls) -> Client:
        config = DarwinConfig.local()
        return cls(config)

    @classmethod
    def from_api_key(cls, api_key: str, datasets_dir: Optional[Path] = None) -> Client:
        config = DarwinConfig.from_api_key_with_defaults(api_key=api_key)
        client = ClientCore(config)  # create a temporary client to get the default team
        token_info = client.get("/users/token_info")
        assert isinstance(token_info, dict)
        default_team: str = token_info["selected_team"]["slug"]
        config.default_team = default_team
        if datasets_dir:
            config.datasets_dir = datasets_dir
        return cls(config)

    @property
    def team(self) -> Team:
        if self._team is None:
            self._team = Team(self)
        return self._team

    # @property
    # def workflows(self) -> WorkflowQuery:
    #     return WorkflowQuery(self, meta_params={"team_slug": self.team.slug})



from __future__ import annotations

from typing import Dict, Generic, Optional, TypeVar

from darwin.future.core.client import ClientCore

R = TypeVar("R")
Param = Dict[str, object]


class MetaBase(Generic[R]):
    """
    A base class for metadata objects. This should only ever be inherited from in meta objects.
    stores metadata parameters used to access the api that are related to the Meta Objects
    but potentially not required for the core object. For example, a dataset object needs
    the team slug to access the api which get's passed down from the team object.

    Attributes:
        _element (R): The element R to which the object is related.
        client (ClientCore): The client used to execute the query.
        meta_params (Dict[str, object]): A dictionary of metadata parameters. This is
            used in conjuction with the Query object to execute related api calls.

    Methods:
        __init__(client: ClientCore, element: R, meta_params: Optional[Param] = None) -> None:
            Initializes a new MetaBase object.
        __repr__() -> str:
            Returns a string representation of the object.

    Examples:
        # Create a MetaBase type that manages a TeamCore object from the API
        class Team(MetaBase[TeamCore]):
            ...
    """

    _element: R
    client: ClientCore

    def __init__(
        self,
        element: R,
        client: ClientCore,
        meta_params: Optional[Param] = None,
    ) -> None:
        self.client = client
        self._element = element
        self.meta_params = meta_params or {}

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self._element})"



from collections import namedtuple
from enum import Enum, auto
from logging import getLogger
from pathlib import Path
from typing import Iterable, List

from darwin.datatypes import Annotation, AnnotationFile, VideoAnnotation
from darwin.exceptions import DarwinException
from darwin.exporter.formats.helpers.yolo_class_builder import (
    ClassIndex,
    build_class_index,
    export_file,
    save_class_index,
)

logger = getLogger(__name__)

CLOSE_VERTICES: bool = False  # Set true if polygons need to be closed


Point = namedtuple("Point", ["x", "y"])


def export(annotation_files: Iterable[AnnotationFile], output_dir: Path) -> None:
    """
    Exports YoloV8 format as segments

    Parameters
    ----------
    annotation_files : Iterable[AnnotationFile]
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new pascalvoc files will be.

    Returns
    -------
    None
    """
    annotation_files = list(annotation_files)

    class_index: ClassIndex = build_class_index(
        # fmt: off
        annotation_files, ["bounding_box", "polygon"]
    )  # fmt: on

    for annotation_file in annotation_files:
        export_file(annotation_file, class_index, output_dir, _build_text)

    save_class_index(class_index, output_dir)


def normalise(value: float, height_or_width: int) -> float:
    """
    Normalises the value to a proportion of the image size

    Parameters
    ----------
    value : float
        The value to be normalised.
    height_or_width : Union[float, int]
        The height or width of the image.

    Returns
    -------
    float
        The normalised value.
    """
    return value / height_or_width


class YoloSegmentedAnnotationType(Enum):
    """
    The YoloV8 annotation types
    """

    UNKNOWN = auto()
    BOUNDING_BOX = auto()
    POLYGON = auto()


def _determine_annotation_type(annotation: Annotation) -> YoloSegmentedAnnotationType:
    """
    Determines the annotation type

    Parameters
    ----------
    annotation : Annotation
        The annotation to be determined.

    Returns
    -------
    YoloSegmentedAnnotationType
        The annotation type.
    """
    type = annotation.annotation_class.annotation_type

    if type == "bounding_box":
        return YoloSegmentedAnnotationType.BOUNDING_BOX
    elif type == "polygon":
        return YoloSegmentedAnnotationType.POLYGON
    else:
        return YoloSegmentedAnnotationType.UNKNOWN


def _handle_bounding_box(
    data: dict, im_w: int, im_h: int, annotation_index: int, points: List[Point]
) -> bool:
    logger.debug(f"Exporting bounding box at index {annotation_index}.")

    try:
        # Create 8 coordinates for the x,y pairs of the 4 corners
        x1, y1, x2, y2, x3, y3, x4, y4, x5, y5 = (
            # top left corner
            data["x"],
            data["y"],
            # top right corner
            (data["x"] + data["w"]),
            (data["y"]),
            # bottom right
            (data["x"] + data["w"]),
            (data["y"] + data["h"]),
            # bottom left
            data["x"],
            (data["y"] + data["h"]),
            # top left again to close the polygon
            data["x"],
            data["y"],
        )

        logger.debug(
            "Coordinates for bounding box: "
            f"({x1}, {y1}), ({x2}, {y2}), "
            f"({x3}, {y3}), ({x4}, {y4}), "
            f"({x5}, {y5})"  # Unsure if we have to close this.
        )

        # Normalize the coordinates to a proportion of the image size
        n_x1 = normalise(x1, im_w)
        n_y1 = normalise(y1, im_h)
        n_x2 = normalise(x2, im_w)
        n_y2 = normalise(y2, im_h)
        n_x3 = normalise(x3, im_w)
        n_y3 = normalise(y3, im_h)
        n_x4 = normalise(x4, im_w)
        n_y4 = normalise(y4, im_h)
        n_x5 = normalise(x5, im_w)
        n_y5 = normalise(y5, im_w)

        logger.debug(
            "Normalized coordinates for bounding box: "
            f"({n_x1}, {n_y1}), ({n_x2}, {n_y2}), "
            f"({n_x3}, {n_y3}), ({n_x4}, {n_y4}), "
            f"({n_x5}, {n_y5})"
        )

        # Add the coordinates to the points list
        points.append(Point(x=n_x1, y=n_y1))
        points.append(Point(x=n_x2, y=n_y2))
        points.append(Point(x=n_x3, y=n_y3))
        points.append(Point(x=n_x4, y=n_y4))

        if CLOSE_VERTICES:
            points.append(Point(x=n_x5, y=n_y5))

    except KeyError as exc:
        logger.warn(
            f"Skipped annotation at index {annotation_index} because an"
            "expected key was not found in the data.",
            exc_info=exc,
        )
        return False

    return True


def _handle_polygon(
    data: dict, im_w: int, im_h: int, annotation_index: int, points: List[Point]
) -> bool:
    logger.debug(f"Exporting polygon at index {annotation_index}.")

    last_point = None
    try:
        if "paths" in data:
            paths_data = data["paths"]
            if len(paths_data) > 1:
                raise DarwinException from ValueError(
                    "Complex polygon detected. The YOLOV8 format only supports simple polygons with a single path."
                )
            path_data = paths_data[0]
        # Continuing to support old versions, in case anything relies on it.
        elif "path" in data:
            path_data = data["path"]
        elif "points" in data:
            path_data = data["points"]
        else:
            raise DarwinException from ValueError("No path data found in annotation.")

        for point_index, point in enumerate(path_data):
            last_point = point_index
            x = point["x"] / im_w
            y = point["y"] / im_h
            points.append(Point(x=x, y=y))

        if CLOSE_VERTICES:
            points.append(points[0])

    except KeyError as exc:
        logger.warn(
            (
                f"Skipped annotation at index {annotation_index} because an"
                "expected key was not found in the data."
                f"Error occured while calculating point at index {last_point}."
                if last_point
                else "Error occured while enumerating points."
            ),
            exc_info=exc,
        )
        return False

    except Exception:
        logger.error(
            f"An unexpected error occured while exporting annotation at index {annotation_index}."
        )

    return True


def _build_text(annotation_file: AnnotationFile, class_index: ClassIndex) -> str:
    """
    Builds the YoloV8 format as segments

    Parameters
    ----------
    annotation_file : AnnotationFile
        The ``AnnotationFile`` to be exported.
    class_index : ClassIndex
        The class index.

    Returns
    -------
    str
        The YoloV8 format as segments
    """
    yolo_lines: List[str] = []

    im_w = annotation_file.image_width
    im_h = annotation_file.image_height

    if not im_w or not im_h:
        raise ValueError(
            "Annotation file has no image width or height. "
            "YoloV8 Segments are encoded as a proportion of height and width. "
            "This file cannot be YoloV8 encoded without image dimensions."
        )

    for annotation_index, annotation in enumerate(annotation_file.annotations):
        # Sanity checks
        if isinstance(annotation, VideoAnnotation):
            logger.warn(
                f"Skipped annotation at index {annotation_index} because video annotations don't contain the needed data."
            )
            continue

        if annotation.data is None:
            logger.warn(
                f"Skipped annotation at index {annotation_index} because it's data fields are empty.'"
            )
            continue

        # Process annotations

        annotation_type = _determine_annotation_type(annotation)
        if annotation_type == YoloSegmentedAnnotationType.UNKNOWN:
            continue

        data = annotation.data
        points: List[Point] = []

        if annotation_type == YoloSegmentedAnnotationType.BOUNDING_BOX:
            bb_success = _handle_bounding_box(
                data, im_w, im_h, annotation_index, points
            )
            if not bb_success:
                continue
        elif annotation_type == YoloSegmentedAnnotationType.POLYGON:
            polygon_success = _handle_polygon(
                data, im_w, im_h, annotation_index, points
            )
            if not polygon_success:
                continue
        else:
            logger.warn(
                f"Skipped annotation at index {annotation_index} because it's annotation type is not supported."
            )
            continue

        if len(points) < 3:
            logger.warn(
                f"Skipped annotation at index {annotation_index} because it "
                "has less than 3 points.  Any valid polygon must have at least"
                " 3 points."
            )
            continue

        # Create the line for the annotation
        yolo_line = f"{class_index[annotation.annotation_class.name]} {' '.join([f'{p.x} {p.y}' for p in points])}"
        yolo_lines.append(yolo_line)
    return "\n".join(yolo_lines) + "\n"



from __future__ import annotations

from typing import Dict, List

from darwin.future.core.team.get_team import get_team_members
from darwin.future.core.types.query import Query, QueryFilter
from darwin.future.meta.objects.team_member import TeamMember


class TeamMemberQuery(Query[TeamMember]):
    """
    TeamMemberQuery object with methods to manage filters, retrieve data,
    and execute filters
    Methods:
    collect: Executes the query and returns the filtered data
    _execute_filter: Executes a filter on a list of objects
    """

    def _collect(self) -> Dict[int, TeamMember]:
        members, exceptions = get_team_members(self.client)
        members_meta = [
            TeamMember(client=self.client, element=member) for member in members
        ]
        if exceptions:
            # TODO: print and or raise exceptions, tbd how we want to handle this
            pass
        if not self.filters:
            self.filters = []
        for filter in self.filters:
            members_meta = self._execute_filter(members_meta, filter)

        return dict(enumerate(members_meta))

    def _execute_filter(
        self, members: List[TeamMember], filter: QueryFilter
    ) -> List[TeamMember]:
        """
        Executes filtering on the local list of members, applying special logic for
        role filtering otherwise calls the parent method for general filtering on the
        values of the members

        Parameters
        ----------
        members : List[TeamMember]
        filter : QueryFilter

        Returns
        -------
        List[TeamMember]: Filtered subset of members
        """
        if filter.name == "role":
            return [
                m
                for m in members
                if m._element is not None and filter.filter_attr(m._element.role.value)
            ]
        else:
            return super()._generic_execute_filter(members, filter)



from typing import Any, Dict, List


import darwin.datatypes as dt

# from darwin.datatypes import PolygonPath, PolygonPaths

DEPRECATION_MESSAGE = """

This function is going to be turned into private. This means that breaking 
changes in its interface and implementation are to be expected. We encourage using ``build_image_annotation`` 
instead of calling this low-level function directly.

"""


def build_image_annotation(
    annotation_file: dt.AnnotationFile, team_name: str
) -> Dict[str, Any]:
    """
    Builds and returns a dictionary with the annotations present in the given file in Darwin v2 format.

    Parameters
    ----------
    annotation_file: AnnotationFile
        File with the image annotations to extract.
        For schema, see: https://darwin-public.s3.eu-west-1.amazonaws.com/darwin_json/2.0/schema.json

    Returns
    -------
    Dict[str, Any]
        A dictionary with the annotations in Darwin v2 format.
    """
    annotations_list: List[Dict[str, Any]] = []

    for annotation in annotation_file.annotations:
        annotation_data = _build_v2_annotation_data(annotation)
        annotations_list.append(annotation_data)

    slots_data = _build_slots_data(annotation_file.slots)
    item = _build_item_data(annotation_file, team_name)
    item["slots"] = slots_data

    return {
        "version": "2.0",
        "schema_ref": "https://darwin-public.s3.eu-west-1.amazonaws.com/darwin_json/2.0/schema.json",
        "item": item,
        "annotations": annotations_list,
    }


def _build_v2_annotation_data(annotation: dt.Annotation) -> Dict[str, Any]:
    annotation_data = {"id": annotation.id, "name": annotation.annotation_class.name}

    if annotation.annotation_class.annotation_type == "bounding_box":
        annotation_data["bounding_box"] = _build_bounding_box_data(annotation.data)
    elif annotation.annotation_class.annotation_type == "tag":
        annotation_data["tag"] = {}
    elif annotation.annotation_class.annotation_type == "polygon":
        polygon_data = _build_polygon_data(annotation.data)
        annotation_data["polygon"] = polygon_data
        annotation_data["bounding_box"] = _build_bounding_box_data(annotation.data)

    return annotation_data


def _build_bounding_box_data(data: Dict[str, Any]) -> Dict[str, Any]:
    if "bounding_box" in data:
        data = data["bounding_box"]
    return {
        "h": data.get("h"),
        "w": data.get("w"),
        "x": data.get("x"),
        "y": data.get("y"),
    }


def _build_polygon_data(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Builds the polygon data for Darwin V2 format from Darwin internal format (looks like V1).

    Parameters
    ----------
    data : Dict[str, Any]
        The original data for the polygon annotation.

    Returns
    -------
    Dict[str, List[List[Dict[str, float]]]]
        The polygon data in the format required for Darwin v2 annotations.
    """
    return {"paths": data["paths"]}


def _build_item_data(
    annotation_file: dt.AnnotationFile, team_name: str
) -> Dict[str, Any]:
    """
    Constructs the 'item' section of the Darwin v2 format annotation.

    Parameters
    ----------
    annotation_file: dt.AnnotationFile
        The AnnotationFile object containing annotation data.

    Returns
    -------
    Dict[str, Any]
        The 'item' section of the Darwin v2 format annotation.
    """
    return {
        "name": annotation_file.filename,
        "path": annotation_file.remote_path or "/",
        "source_info": {
            "dataset": {
                "name": annotation_file.dataset_name,
                "slug": (
                    annotation_file.dataset_name.lower().replace(" ", "-")
                    if annotation_file.dataset_name
                    else None
                ),
            },
            "item_id": annotation_file.item_id,
            "team": {
                "name": team_name,
                "slug": team_name.lower().replace(" ", "-"),
            },
            "workview_url": annotation_file.workview_url,
        },
    }


def _build_slots_data(slots: List[dt.Slot]) -> List[Dict[str, Any]]:
    """
    Constructs the 'slots' data for the Darwin v2 format annotation.

    Parameters
    ----------
    slots: List[Slot]
        A list of Slot objects from the AnnotationFile.

    Returns
    -------
    List[Dict[str, Any]]
        The 'slots' data for the Darwin v2 format annotation.
    """
    slots_data = []
    for slot in slots:
        slot_data = {
            "type": slot.type,
            "slot_name": slot.name,
            "width": slot.width,
            "height": slot.height,
            "thumbnail_url": slot.thumbnail_url,
            "source_files": slot.source_files,
        }
        slots_data.append(slot_data)

    return slots_data



from darwin.future.core.workflows.get_workflow import get_workflow
from darwin.future.core.workflows.get_workflows import get_workflows
from darwin.future.core.workflows.list_workflows import list_workflows



import responses
from responses import json_params_matcher
from responses.matchers import query_param_matcher

from darwin.future.meta.objects.workflow import Workflow
from darwin.future.tests.meta.objects.fixtures import *


def test_item_set_stage_with_stage_object(
    workflow: Workflow,
    base_single_workflow_object: list,
    base_items_json_response: dict,
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            workflow.client.config.api_endpoint
            + f"v2/teams/default-team/workflows/{workflow.id}",
            status=200,
            json=base_single_workflow_object,
        )
        stages = workflow.stages.collect()
        rsps.add(
            rsps.GET,
            workflow.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "0",
                        "page[size]": "500",
                        "dataset_ids": "101",
                    }
                )
            ],
            status=200,
            json=base_items_json_response,
        )
        items = workflow.items.collect()
        item = items[0]
        stage = stages[0]

        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        workflow_id = str(workflow.id)
        stage_id = str(stage.id)
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/stage",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "stage_id": stage_id,
                        "workflow_id": workflow_id,
                    }
                )
            ],
            json={},
        )
        item.set_stage(stage)



import pytest
from pydantic import ValidationError

from darwin.future.data_objects.sorting import SortingMethods


def test_sorting_methods_all_fields_none():
    with pytest.raises(ValidationError):
        SortingMethods()


def test_sorting_methods_one_field_set():
    sorting = SortingMethods(accuracy="asc")
    assert sorting.accuracy == "asc"


def test_sorting_methods_multiple_fields_set():
    sorting = SortingMethods(accuracy="asc", byte_size="desc")
    assert sorting.accuracy == "asc"
    assert sorting.byte_size == "desc"


def test_sorting_methods_invalid_value():
    with pytest.raises(ValidationError):
        SortingMethods(accuracy="invalid")



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar

from pydantic import field_validator
from typing_extensions import Self

from darwin.future.core.client import ClientCore
from darwin.future.core.types.common import Stringable
from darwin.future.data_objects.advanced_filters import GroupFilter, SubjectFilter
from darwin.future.data_objects.page import Page
from darwin.future.exceptions import (
    InvalidQueryFilter,
    InvalidQueryModifier,
    MoreThanOneResultFound,
    ResultsNotFound,
)
from darwin.future.meta.objects.base import MetaBase
from darwin.future.pydantic_base import DefaultDarwin

T = TypeVar("T", bound=MetaBase)
R = TypeVar("R", bound=DefaultDarwin)
Param = Dict[str, Any]  # type: ignore


class Modifier(Enum):
    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_EQUAL = ">="
    LESS_EQUAL = "<="
    NOT_EQUAL = "!="
    CONTAINS = "contains"


class QueryFilter(DefaultDarwin):
    """
    Basic query filter with a name and a parameter
    Modifiers are for client side filtering only, and are not passed to the API
    Attributes
    ----------
    name: str
    param: str
    modifier: Optional[Modifiers]: Optional modifier to apply to the filter
    """

    name: str
    param: str
    modifier: Optional[Modifier] = None

    @field_validator("param")
    def validate_param(cls, v: Stringable) -> str:
        return str(v)

    def filter_attr(self, attr: Any) -> bool:  # type: ignore
        caster: Callable[[str], Any] = type(attr)  # type: ignore
        param = caster(
            self.param
        )  # attempt to cast the param to the type of the attribute
        if self.modifier is None:
            return attr == param
        elif self.modifier == Modifier.GREATER_EQUAL:
            return attr >= param
        elif self.modifier == Modifier.GREATER_THAN:
            return attr > param
        elif self.modifier == Modifier.LESS_EQUAL:
            return attr <= param
        elif self.modifier == Modifier.LESS_THAN:
            return attr < param
        elif self.modifier == Modifier.NOT_EQUAL:
            return attr != param
        elif self.modifier == Modifier.CONTAINS:
            return param in attr
        else:
            raise InvalidQueryModifier(f"Unknown modifier {self.modifier}")

    @classmethod
    def _from_dict(cls, d: Dict[str, Any]) -> QueryFilter:  # type: ignore
        if "name" not in d or "param" not in d:
            raise InvalidQueryFilter(
                "args must be a QueryFilter or a dict with 'name' and 'param' keys,"
                f" got {d}"
            )
        modifier = Modifier(d["modifier"]) if "modifier" in d else None
        return QueryFilter(name=d["name"], param=str(d["param"]), modifier=modifier)

    @classmethod
    def _from_args(cls, *args: object, **kwargs: str) -> List[QueryFilter]:
        filters = []
        for arg in args:
            filters.append(cls._from_arg(arg))
        for key, value in kwargs.items():
            filters.append(cls._from_kwarg(key, value))
        return filters

    @classmethod
    def _from_arg(cls, arg: object) -> QueryFilter:
        if isinstance(arg, QueryFilter):
            return arg
        elif isinstance(arg, dict):
            return cls._from_dict(arg)
        else:
            raise InvalidQueryFilter(
                "args must be a QueryFilter or a dict with 'name' and 'param' keys,"
                f" got {arg}"
            )

    @classmethod
    def _from_kwarg(cls, key: str, value: str) -> QueryFilter:
        if ":" in value:
            modifier_str, value = value.split(":", 1)
            modifier = Modifier(modifier_str)
        else:
            modifier = None
        return QueryFilter(name=key, param=value, modifier=modifier)

    def to_dict(self, ignore_modifier: bool = True) -> Dict[str, str]:
        d = {self.name: self.param}
        if self.modifier is not None and not ignore_modifier:
            d["modifier"] = self.modifier.value
        return d


class Query(Generic[T], ABC):
    """
    A basic Query object with methods to manage filters. This is an abstract class not
    meant to be used directly. Use a subclass instead, like DatasetQuery.
    This class will lazy load results and cache them internally, and allows for filtering
    of the objects locally by default. To execute the query, call the collect() method,
    or iterate over the query object.

    Attributes:
        meta_params (dict): A dictionary of metadata parameters.
        client (ClientCore): The client used to execute the query.
        filters (List[QueryFilter]): A list of QueryFilter objects used to filter the query results.
        results (List[T]): A list of query results, cached internally for iterable access.
        _changed_since_last (bool): A boolean indicating whether the query has changed since the last execution.

    Methods:
        filter(name: str, param: str, modifier: Optional[Modifier] = None) -> Self:
            Adds a filter to the query object and returns a new query object.
        where(name: str, param: str, modifier: Optional[Modifier] = None) -> Self:
            Applies a filter on the query object and returns a new query object.
        first() -> Optional[T]:
            Returns the first result of the query. Raises an exception if no results are found.
        collect() -> List[T]:
            Executes the query on the client and returns the results. Raises an exception if no results are found.
        _generic_execute_filter(objects: List[T], filter_: QueryFilter) -> List[T]:
            Executes a filter on a list of objects. Locally by default, but can be overwritten by subclasses.

    Examples:
        # Create a query object
        # DatasetQuery is linked to the object it returns, Dataset, and is iterable
        # overwrite the _collect() method to execute insantiate this object
        Class DatasetQuery(Query[Dataset]):
            ...

        # Intended usage via chaining
        # where client.team.datasets returns a DatasetQuery object and can be chained
        # further with multiple where calls before collecting
        datasets = client.team.datasets.where(...).where(...).collect()
    """

    def __init__(
        self,
        client: ClientCore,
        filters: Optional[List[QueryFilter]] = None,
        meta_params: Optional[Param] = None,
    ):
        self.meta_params: dict = meta_params or {}
        self.client = client
        self.filters = filters or []
        self.results: dict[int, T] = {}
        self._changed_since_last: bool = False

    def filter(self, filter: QueryFilter) -> Self:
        return self + filter

    def __add__(self, filter: QueryFilter) -> Self:
        self._changed_since_last = True
        self.filters.append(filter)
        return self

    def __sub__(self, filter: QueryFilter) -> Self:
        self._changed_since_last = True
        return self.__class__(
            self.client,
            filters=[f for f in self.filters if f != filter],
            meta_params=self.meta_params,
        )

    def __iadd__(self, filter: QueryFilter) -> Self:
        self.filters.append(filter)
        self._changed_since_last = True
        return self

    def __isub__(self, filter: QueryFilter) -> Self:
        self.filters = [f for f in self.filters if f != filter]
        self._changed_since_last = True
        return self

    def __len__(self) -> int:
        if not self.results:
            self.results = {**self.results, **self._collect()}
        return len(self.results)

    def __iter__(self) -> Self:
        self.n = 0
        return self

    def __next__(self) -> T:
        if not self.results:
            self.collect()
        if self.n < len(self.results):
            result = self.results[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def __getitem__(self, index: int) -> T:
        if not self.results:
            self.results = {**self.results, **self._collect()}
        return self.results[index]

    def __setitem__(self, index: int, value: T) -> None:
        if not self.results:
            self.results = {**self.results, **self._collect()}
        self.results[index] = value

    def where(self, *args: object, **kwargs: str) -> Self:
        filters = QueryFilter._from_args(*args, **kwargs)
        for item in filters:
            self += item
        self._changed_since_last = True
        return self

    def collect(self, force: bool = False) -> List[T]:
        if force or self._changed_since_last:
            self.results = {}
        self.results = {**self.results, **self._collect()}
        self._changed_since_last = False
        return self._unwrap(self.results)

    def _unwrap(self, results: Dict[int, T]) -> List[T]:
        return list(results.values())

    @abstractmethod
    def _collect(self) -> Dict[int, T]:
        raise NotImplementedError("Not implemented")

    def collect_one(self) -> T:
        if not self.results:
            self.results = {**self.results, **self._collect()}
        if len(self.results) == 0:
            raise ResultsNotFound("No results found")
        if len(self.results) > 1:
            raise MoreThanOneResultFound("More than one result found")
        return self.results[0]

    def first(self) -> T:
        if not self.results:
            self.results = {**self.results, **self._collect()}
        if len(self.results) == 0:
            raise ResultsNotFound("No results found")

        return self.results[0]

    def _generic_execute_filter(self, objects: List[T], filter: QueryFilter) -> List[T]:
        return [
            m for m in objects if filter.filter_attr(getattr(m._element, filter.name))
        ]


class PaginatedQuery(Query[T]):
    def __init__(
        self,
        client: ClientCore,
        filters: List[QueryFilter] | None = None,
        meta_params: Param | None = None,
        page: Page | None = None,
    ):
        super().__init__(client, filters, meta_params)
        self._advanced_filters: GroupFilter | SubjectFilter | None = None
        self.page = page or Page()
        self.completed = False

    def collect(self, force: bool = False) -> List[T]:
        if force or self._changed_since_last:
            self.page = Page()
            self.completed = False
        if self.completed:
            return self._unwrap(self.results)
        new_results = self._collect()
        self.results = {**self.results, **new_results}
        if len(new_results) < self.page.size or len(new_results) == 0:
            self.completed = True
        else:
            self.page.increment()
        return self._unwrap(self.results)

    def collect_all(self, force: bool = False) -> List[T]:
        if force:
            self.page = Page()
            self.completed = False
            self.results = {}
        while not self.completed:
            self.collect()
        return self._unwrap(self.results)

    def __getitem__(self, index: int) -> T:
        if index not in self.results:
            temp_page = self.page
            self.page = self.page.get_required_page(index)
            self.collect()
            self.page = temp_page
        return super().__getitem__(index)

    def __next__(self) -> T:
        if not self.completed and self.n not in self.results:
            self.collect()
        if self.completed and self.n not in self.results:
            raise StopIteration
        result = self.results[self.n]
        self.n += 1
        return result

    def __len__(self) -> int:
        if not self.completed:
            self.collect_all()
        return len(self.results)



import os
import sys
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

import numpy as np
import torch
from numpy.typing import ArrayLike
from upolygon import draw_polygon

from darwin.cli_functions import _error, _load_client
from darwin.dataset.identifier import DatasetIdentifier
from darwin.datatypes import Segment


def flatten_masks_by_category(masks: torch.Tensor, cats: List[int]) -> torch.Tensor:
    """
    Takes a list of masks and flattens into a single mask output with category id's overlaid into one tensor.
    Overlapping sections of masks are replaced with the top most annotation in that position
    Parameters
    ----------
    masks : torch.Tensor
        lists of masks with shape [x, image_height, image_width] where x is the number of categories
    cats : List[int]
        int list of category id's with len(x)
    Returns
    -------
    torch.Tensor
        Flattened mask of category id's
    """
    assert isinstance(masks, torch.Tensor)
    assert isinstance(cats, List)
    assert masks.shape[0] == len(cats)
    order_of_polygons = list(range(1, len(cats) + 1))
    polygon_mapping = {order: cat for cat, order in zip(cats, order_of_polygons)}
    BACKGROUND: int = 0
    polygon_mapping[BACKGROUND] = 0
    # Uses matrix multiplication here with `masks` being a binary array of same dimensions as image
    # and polygon orders being overlaid onto the relevant mask
    order_tensor = torch.as_tensor(order_of_polygons, dtype=masks.dtype)
    flattened, _ = (masks * order_tensor[:, None, None]).max(dim=0)
    # The mask is now flattened in order of the polygons but needs to be converted back to the categories
    # vectorize the dictionary to return the original category id's
    mapped = np.vectorize(polygon_mapping.__getitem__)(flattened)
    return torch.as_tensor(mapped, dtype=masks.dtype)


def convert_segmentation_to_mask(
    segmentations: List[Segment], height: int, width: int
) -> torch.Tensor:
    """
    Converts a polygon represented as a sequence of coordinates into a mask.

    Parameters
    ----------
    segmentations : List[Segment]
        List of float values -> ``[[x11, y11, x12, y12], ..., [xn1, yn1, xn2, yn2]]``.
    height : int
        Image's height.
    width : int
        Image's width.

    Returns
    -------
    torch.tensor
        A ``Tensor`` representing a segmentation mask.
    """
    if not segmentations:
        return torch.zeros((0, height, width), dtype=torch.uint8)

    masks = []
    for contour in segmentations:
        mask = torch.zeros((height, width)).numpy().astype(np.uint8)
        masks.append(torch.from_numpy(np.asarray(draw_polygon(mask, contour, 1))))
    return torch.stack(masks)


def polygon_area(x: ArrayLike, y: ArrayLike) -> float:
    """
    Returns the area of the input polygon, represented by two numpy arrays for x and y coordinates.

    Parameters
    ----------
    x : np.ndarray
        Numpy array for x coordinates.
    y : np.ndarray
        Numpy array for y coordinates.

    Returns
    -------
    float
        The area of the polygon.
    """
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def collate_fn(batch: Iterable[Tuple]) -> Tuple:
    """
    Aggregates the given ``Iterable`` (usually a ``List``) of tuples into a ``Tuple`` of Lists.

    Parameters
    ----------
    batch : Iterable[Tuple]
        Batch to collate.

    Returns
    -------
    Tuple
        The ``Iterable`` of Tupled aggregated into a ``Tuple``.
    """
    return tuple(zip(*batch))


def detectron2_register_dataset(
    dataset: str,
    release_name: Optional[str] = "latest",
    partition: Optional[str] = None,
    split: Optional[str] = "default",
    split_type: Optional[str] = "stratified",
    evaluator_type: Optional[str] = None,
) -> str:
    """
    Registers a local Darwin-formatted dataset in Detectron2.

    Parameters
    ----------
    dataset : str
        Dataset slug.
    release_name : Optional[str], default: "latest"
        Version of the dataset.
    partition : Optional[str], default: None
        Selects one of the partitions ``["train", "val", "test"]``.
    split : Optional[str], default: "default"
        Selects the split that defines the percentages used.
    split_type : Optional[str], default: "stratified"
        Heuristic used to do the split ``["random", "stratified"]``.
    evaluator_type : Optional[str], default: None
        Evaluator to be used in the val and test sets.

    Returns
    -------
    str
        The name of the registered dataset in the format of ``{dataset-name}_{partition}``.
    """
    try:
        from detectron2.data import DatasetCatalog, MetadataCatalog
    except ImportError:
        print("Detectron2 not found.")
        sys.exit(1)
    from darwin.dataset.utils import get_annotations, get_classes

    dataset_path: Optional[Path] = None
    if os.path.isdir(dataset):
        dataset_path = Path(dataset)
    else:
        identifier = DatasetIdentifier.parse(dataset)
        if identifier.version:
            release_name = identifier.version

        client = _load_client(offline=True)
        dataset_path = None
        for path in client.list_local_datasets(team_slug=identifier.team_slug):
            if identifier.dataset_slug == path.name:
                dataset_path = path

        if not dataset_path:
            _error(
                f"Dataset '{identifier.dataset_slug}' does not exist locally. "
                f"Use 'darwin dataset remote' to see all the available datasets, "
                f"and 'darwin dataset pull' to pull them."
            )

    catalog_name = f"darwin_{dataset_path.name}"
    if partition:
        catalog_name += f"_{partition}"

    classes = get_classes(
        dataset_path=dataset_path, release_name=release_name, annotation_type="polygon"
    )

    DatasetCatalog.register(
        catalog_name,
        lambda partition=partition: list(
            get_annotations(
                dataset_path,
                partition=partition,
                split=split,
                split_type=split_type,
                release_name=release_name,
                annotation_type="polygon",
                annotation_format="coco",
                ignore_inconsistent_examples=True,
            )
        ),
    )
    MetadataCatalog.get(catalog_name).set(thing_classes=classes)
    if evaluator_type:
        MetadataCatalog.get(catalog_name).set(evaluator_type=evaluator_type)
    return catalog_name


def clamp_bbox_to_image_size(annotations, img_width, img_height, format="xywh"):
    """
    Clamps bounding boxes in annotations to the given image dimensions.

    :param annotations: Dictionary containing bounding box coordinates in 'boxes' key.
    :param img_width: Width of the image.
    :param img_height: Height of the image.
    :param format: Format of the bounding boxes, either "xywh" or "xyxy".
    :return: Annotations with clamped bounding boxes.

    The function modifies the input annotations dictionary to clamp the bounding box coordinates
    based on the specified format, ensuring they lie within the image dimensions.
    """
    boxes = annotations["boxes"]

    if format == "xyxy":
        boxes[:, 0::2].clamp_(min=0, max=img_width - 1)
        boxes[:, 1::2].clamp_(min=0, max=img_height - 1)

    elif format == "xywh":
        # First, clamp the x and y coordinates
        boxes[:, 0].clamp_(min=0, max=img_width - 1)
        boxes[:, 1].clamp_(min=0, max=img_height - 1)
        # Then, clamp the width and height
        boxes[:, 2].clamp_(
            min=torch.tensor(0), max=img_width - boxes[:, 0] - 1
        )  # -1 since we images are zero-indexed
        boxes[:, 3].clamp_(
            min=torch.tensor(0), max=img_height - boxes[:, 1] - 1
        )  # -1 since we images are zero-indexed
    else:
        raise ValueError(f"Unsupported bounding box format: {format}")

    annotations["boxes"] = boxes
    return annotations
